open Robot
open World
open Scanf
open Printf
open Str
open String
open Unix
open Sys
open Misc
open Miscworld

class game = 
   fun inc outc -> 
      let tmpw = new world inc in
      let str = input_line inc in
      let (mid, mcap, mmoney) = sscanf str "%i %i %i" (fun x y z -> (x,y,z)) 
      in object(self)
	    val map = tmpw
	    val my_id = mid			
	    val me = new my_robot mid mcap mmoney
	    val capacity = mcap
	    val money = mmoney
	    val in_c = inc
	    val out_c = outc (*Pervasives.stdout*)
	    val mutable turn = 0
			  
	    method private parse_reply () =
	       let str = (input_line in_c) in		  
	       let botsops = Str.split (regexp "#") (sub str 1 (String.length str - 1)) in
		  List.iter (fun str ->
				let list = Str.split (regexp " ") str in
				let id = int_of_string(List.hd list) in
				let rec recfun (l:string list) =
				      match l with
					  [] -> ()					     
					| "P"::tl -> 
					     let tmp = ref (tl) in
					     let n = ref (my_int_of_string(List.hd !tmp)) in
						while (!n != -1) do
						   map#update id (Pick [!n]);
						   tmp := (List.tl !tmp);
						   if !tmp != [] then
						      n := my_int_of_string (List.hd !tmp)
						   else
						      n := -1
						done;
						recfun !tmp;
					| "D"::tl ->
					     let tmp = ref (tl) in
					     let n = ref (my_int_of_string (List.hd !tmp)) in
						while (!n != -1) do
						   map#update id (Drop [!n]);
						   tmp := (List.tl !tmp);
						   if !tmp != [] then
						      n := my_int_of_string (List.hd !tmp)
						   else
						      n := -1
						done;
						recfun !tmp;
						
					| "X"::tl -> self#add_bot id { 
					     x = int_of_string(List.hd tl) - 1; 
					     y = int_of_string(List.hd (List.tl (List.tl tl))) - 1 };
					     recfun (List.tl (List.tl (List.tl tl)))
					| hd::tl -> 
					     map#update id (Move(direction_of_string (hd)));
					     recfun tl;
				in
				   recfun (List.tl list))
		     botsops 


	    method add_bot id coord =
	       (*printf "Adding bot %i with coords (%i,%i)" id coord.x coord.y;
	       print_newline();*)
	       if id = my_id then (
		  me#set_coords coord;
		  map#add_bot (me :> robot) coord;
	       ) else
		  let bot = new robot id coord money capacity in
		     map#add_bot bot coord;
		          
	    method private parse_packs () =
	       let str = (input_line in_c) in
		  if str = "" then 
		     ()
		  else (
		     (*printf "Server indicate %s" str;
		     print_newline ();*)
		     let list = Array.of_list (List.map (fun x -> int_of_string(x)) 
						  (Str.split (regexp " ") str)) in
			for i = 0 to ((Array.length list) / 4) - 1 do
			   map#drop_pack
			      {
				 id = list.(i * 4);
				 weight = list.(i * 4 + 3);
				 dst = {
				    x = list.(i * 4 + 1) - 1; 
				    y = list.(i * 4 + 2) - 1; 
				 };
				 pos = {x=0;y=0};
			      }
			      me # coords;
			done
		  )		 	      

	    method start () =
	       map#build_matrix;
	       (*map#print_zones ();*)
	       self#parse_reply ();       
	       self#parse_packs ();
	       self#play ();

	    method play () = 
	       while (true) do
		  ignore(
		     turn <- turn + 1;
		     self# act ();
		     self# parse_reply ();
		     self# parse_packs ();
		     (* print_string ("Turn: " ^ (string_of_int turn));
			print_newline ();

		     map#print ();
		       print_newline ();
		       
		     print_endline ("Playing");
		     
		     print_endline ("Played");*)
		  );
	       done
	       
	    method action =
	       if me#in_transfer then (
		  (* delivering or moving toward a package *)
		  map#continue_move me
	       ) else (
		  (* searching *)
		  if turn < 20 then (
		     let home = List.hd map#homes in
		     let p = map#find_way me#coords home in
			p#rev;
			me#set_path p;
			map#continue_move me;
		  ) else (
		     if me#has_package then (
			if me#has_path then (
			   { op = Drop(List.map (fun (id,pack) -> id) me#good_packages); 
			     bid = 10 }			   			   
			) else (
			   let p = map#find_way me#coords (snd (me#packages#first)).dst in
			      p#rev;
			      me#set_path p;
			      map#continue_move me;
			)
		     ) else (
			(* must find packs *)
			me#clear_path;
			map#pick_move me;
		     )
		  ) 
	       )
		  
	    method act () =
	       let print_ids = (fun x -> output_string out_c " "; output_string out_c (string_of_int x);) 
	       and ac = self#action in
		  output_string out_c (string_of_int ac.bid);
		  (
		     match ac.op with
			 Pick ids -> output_string out_c " Pick";
			    List.iter print_ids ids;
		       | Drop ids -> output_string out_c " Drop";
			    List.iter print_ids ids;
		       | Move dir -> output_string out_c " Move ";
			    output_char out_c (char_of_direction dir);
		  );
		  output_string out_c "\n";
		  flush out_c;		  
	 end;;  

let (inc, outc) = open_connection (ADDR_INET(inet_addr_of_string(argv.(1)), int_of_string(argv.(2))));;

output_string outc "Player\n";;
flush outc;;


let game = new game inc outc;;

try 
   game#start ();
with End_of_file ->
   exit 0
