open Unix
open Sys
open Scanf
open Printf
open String
open Str
open Misc
open Assoclist
open Miscworld

type op = 
     Move of direction
   | Pick of int list
   | Drop of int list;;

class robot mid coords money capacity = object(self)
   val id = (mid:int)
   val mutable c = coords
   val mutable money = (money:int)
   val mutable cap = (capacity:int)
   val packs = (new assoclist:(int, package) assoclist)
		  
   method id = id
		  
   method x = c.x
   method y = c.y
		 
   method coords = c
		      
   method move dir = 
      c <- move_dir c dir;

   method capacity = cap
			
   method money = money
		     
   method has_package = 
      not (packs#empty);
	    
   method packages = packs
			
   method drop id = 
      let p = packs#get_rem id in
	 cap <- cap + p.weight;
	 p;
	 
   method pick pack = 
      cap <- cap - pack.weight;
      packs#add pack.id pack;      
      
end;;


type action = {
   op: op;
   bid: int;
}

class my_robot mid mcap mmoney = object(self)
   inherit robot mid {x=0;y=0} 0 0

   val mutable path = (None:path option)

   method random_move =
      { op = Move(East); bid = 1 }
      
   method set_coords new_coords = 
      c <- new_coords

   method in_transfer = 
      match path with
	  Some p ->
	     (p#dst.x != c.x) || (p#dst.y != c.y);
	| None -> false
	     
   method has_arrived =
      match path with
	  Some p ->
	     (p#dst.x == c.x) && (p#dst.y == c.y);
	| None -> raise (Failure "No path set");

   method continue_transfer = 
      {op = Move(North); bid = 1}
      
   method move_to v = 
      if v.x > c.x then
	 if v.y > c.y then 
	    {op = Move(North); bid = 1}
	 else
	    {op = Move(South); bid = 1}
      else
	 {op = Move(East); bid = 1}
     
   method set_path p =
      path <- Some p;

   method clear_path = 
      path <- None

   method has_path =
      match path with
	  Some p -> true
	| None -> false
      
   method path = 
      match path with
	  Some p -> p
	| None -> raise (Failure "No path set")

   method good_packages =
      List.filter 
	 (fun (x,i) ->
	     if i.dst = c then
		true
	     else
		false)
	 packs#list;
end;;
