open Robot
open Scanf
open Printf
open Assoclist
open Stack
open Miscworld

type good_square = {
   mutable bot: robot option;
   mutable packs : (int, package) assoclist;
   mutable mark: bool;
}

type square =
     Open of good_square 
   | Home of good_square
   | Wall
   | Water;;

let good_pred s = match s with
     Open(a) -> true
   | Home(a) -> true
   | _ -> false;;

let pack_pred s = match s with
     Open(a) -> not(a.packs#empty)
   | Home(a) -> not(a.packs#empty)
   | _ -> false;;

		   
class world = 
   fun in_ch ->
      let (w, h) = sscanf (input_line in_ch) "%i %i" (fun x y -> (x, y)) in
      let (arr, homeslist) =  
	 let str = ref ""  and array = Array.make_matrix w h Water and home_bases = ref [] in
	    for i = 0 to (h - 1) do
	       str := (input_line in_ch);
	       for j = 0 to w - 1 do
		  array.(j).(i) <- match !str.[j] with
		      '.' -> Open({bot = None; packs = new assoclist; mark = false;})
		    | '@' -> 
			 home_bases := {x = j; y = i}::!home_bases;
			 Home({bot = None; packs = new assoclist;mark = false;})
		    | '~' -> Water
		    | '#' -> Wall
		    | _ as c -> raise (Failure ("Unknown tile " ^ (Char.escaped c)))
	       done
	    done;
	    (array,!home_bases)
	    
      in object(self)
	    val map = arr
	    val bots = (Hashtbl.create 6:(int, robot) Hashtbl.t)
	    val packs = Hashtbl.create 10000
	    val homes = homeslist
	    val mutable zones = ([]:zone list)

	    method find_near co ray predicate =
	       let rec rec_find_near = fun c r pred ->
		  let verts = [move co (-1) 0; move co 0 1; move co 1 0; move co 0 (-1)] in
		     List.fold_left (fun list c -> 
					let res = 
					   if predicate (self#at c) then
					      c::list
					   else
					      list 
					in
					   if ray > 1 then
					      res @ (rec_find_near co (ray - 1) predicate)
					   else
					      res) [] verts
	       in
		  rec_find_near co ray predicate

	    method at c =
	       map.(c.x).(c.y);
		  
	    method add_bot bot coord = 
	       Hashtbl.add bots bot#id bot;
	       match self#at coord with
		   Home h -> 
		      h.bot <- Some bot;
		 | Open o -> 
		      o.bot <- Some bot;
		 | _ -> raise (Failure "Cannot add bot to bad square")
	            		      		      
	    method update bot op = 
	       try
		  let b = Hashtbl.find bots bot in
		     (match op with 
			  Move(dir) -> 
			     (match self#at b#coords with
				  Open(s) | Home(s) ->
				     s.bot <- None
				| _ -> raise (Failure "Invalid state in map")
			     );
			     b#move dir;
			     
			     (match self#at b#coords with
				 Open(s) | Home(s) ->
				    s.bot <- Some b
				| _ -> raise (Failure "Cannot move bot to bad square")
			     );
			     
			| Pick(id) -> 
			     b#pick (self#pick_pack (List.hd id));
			| Drop(id) ->
			     let p = b#drop (List.hd id) in
				self#drop_pack p p.pos;
		     )
	       with Not_found -> ()
		  
	    method pick_pack id = 
	       let p = Hashtbl.find packs id in
		  match self#at p.pos with 
		      Home(s) -> s.packs#get_rem id
		    | Open(s) -> s.packs#get_rem id
		    | _ -> raise (Failure "Trying to pick pack from bad square")
			 
	    method drop_pack pack coord =
	       pack.pos <- coord;
	       Hashtbl.add packs pack.id pack;
	       match self#at coord with 
		   Home(s) -> s.packs#add pack.id pack
		 | Open(s) -> s.packs#add pack.id pack
		 | _ -> raise (Failure "Trying to drop pack on bad square")
		      
	    method homes = homes

	    method build_matrix = 
	       let tmpx = ref (new interval (-1) (-1))
	       and lastline = (new assoclist:(int, zone ref) assoclist)
	       and newline = (new assoclist:(int, zone ref) assoclist)
	       in
	       let try_merge j = 
		  if !tmpx#fst != -1 then (				  
		     try
			(*printf "try_merge for (%i,%i) %i\n" !tmpx#fst !tmpx#lst j;*)
			(* merge zones *)
			let adj = lastline#get !tmpx#fst in
			   if !adj#w#lst = !tmpx#lst then (
			      (*printf "Merging: %s with %i,%i\n" !adj#to_string !tmpx#fst !tmpx#lst;*)
			      !adj#h#extend j;
			      newline#add !tmpx#fst adj;
			   )
			   else (
			      let z = ref (new zone !tmpx (new interval j j)) in
				 List.iter (fun (id,zo) -> 
					       if !zo#intersect_horiz !tmpx then (
						  !zo#add_relative z;
						  !z#add_relative zo;
					       )) lastline#list;
				 
				 zones <- !adj :: zones;

				 newline#add !tmpx#fst z;
			   )
		     with Not_found ->
			let z = (ref (new zone !tmpx (new interval j j))) in
			   List.iter (fun (id,zo) -> 
					 if !zo#intersect_horiz !tmpx then
					    (
					       !zo#add_relative z;
					       !z#add_relative zo;
					    )) lastline#list;
			   
			   newline#add !tmpx#fst z;
		  )
		  else ();
	       in
		  for j = 0 to Array.length map.(0) - 1 do
		     for i = 0 to Array.length map - 1 do
			match map.(i).(j) with
			    Open(sq) -> 
			       !tmpx#extend i
			  | Home(sq) ->
			       !tmpx#extend i
			  | _ -> 
			       try_merge j;
			       tmpx := new interval (-1) (-1);
		     done;
		     try_merge j;
		     tmpx := new interval (-1) (-1);
		     
		     List.iter (fun (x,y) ->
				   if !y#h#lst != j then (
				      zones <- !y :: zones;
				   );
			       )
			lastline#list;

		     lastline#set_list newline#list;
		     newline#clear;
		  done;
		  zones <- zones @ (List.map (fun (a, c) -> !c) lastline#list);
		  
	    method find_way src dst =
	       let rec find_path (curpath:path) next dest =
		  if curpath#contains next then
		     raise Cycle;
		  
		  (*printf "Find_path next=%s dest=%s\n" !next#to_string !dest#to_string;*)
 
		  curpath#push next;
		  
		  if next = dest then
		     curpath
		  else
		     let pathes = List.map 
				     (fun (x,y) -> 
					 try
					    Some (find_path curpath#copy x dest)
					 with 
					     Cycle -> 
						None
					   | No_path ->
						None
				     ) 
				     !next#relatives 
		     in
			match (find_shortest pathes) with
			    Some p -> p
			  | None -> 			       
			       raise No_path
	       in
		  find_path (create_path src dst) (ref (List.find (zone_includep src) zones))
		     (ref (List.find (zone_includep dst) zones));
		  
	    method continue_move (me:my_robot) =
	       let p = me#path in
	       let cur = p#pop in
		  if !cur#incl p#dst then		
		     let ori = direction_of_move me#coords p#dst in
			p#push cur;
			{ op = Move(ori); bid = 10 }
		  else (
		     let nz = p#top in
		     let ori = !cur#orientation nz in
		     let tmpc = move_dir me#coords ori in			
			if !cur#incl tmpc then (
			   p#push cur;
			   { op = Move(ori); bid = 10 }
			) else if !nz#incl tmpc then
			   { op = Move(ori); bid = 10 }
			else (
			   p#push cur;
			   (* The next square is bad, approach the new zone *)
			   match ori with 
			       North | South ->
				  if !nz#w#lst > me#x then
				     { op = Move(East); bid = 10 }
				  else
				     { op = Move(West); bid = 10 }
			     | East | West ->
				  if !nz#h#lst > me#y then
				     { op = Move(North); bid = 10 }
				  else
				     { op = Move(South); bid = 10 }
			)
		  )
		     
	    method pick_move (me:my_robot) =
	       let sq = 
		  match self#at me#coords with
		      Home(s) | Open(s) ->
			 s
		    | _ -> 
			 raise (Failure "Bad square selected");
	       in		  
		  if not sq.packs#empty then		     
		     { op = Pick ([fst (sq.packs#first)]);
				  bid = 10 }
		  else
		     let home = List.hd homes in
		     let p = self#find_way me#coords home in
			p#rev;
			me#set_path p;
			self#continue_move me;
			
		     
	    method print_zones () =
	       List.iter (fun (x:zone) -> 
			     print_string x#to_string;
			     print_newline()) 
		  zones;
	       print_newline ();

	    method print () = 
	       let height = Array.length map.(0) - 1 in
		  print_string "    ";
		  for i = 0 to Array.length map - 1 do
		     print_int (i / 10)
		  done;
		  print_newline ();
		  print_string "    ";
		  for i = 0 to Array.length map - 1 do
		     print_int (i mod 10)
		  done;
		  for j = 0 to height do
		     print_newline ();
		     printf "%3i " (height - j);
		     for i = 0 to Array.length map - 1 do
			match map.(i).(height - j) with
			    Water -> print_char '~'
			  | Wall -> print_char '#'
			  | Home(sq) -> 
			       begin
				  match sq.bot with
				      None -> print_char '@'
				    | Some b -> print_char 'b'
			       end
			  | Open(sq) ->
			       begin
				  match sq.bot with
				      None -> print_char '.'
				    | Some b -> print_char 'b'
			       end
		     done
		  done
	 end;;

